from pathlib import Path
from typing import Any

from beartype import beartype
from hypothesis.strategies import SearchStrategy
from sqlalchemy import MetaData
from sqlalchemy.engine import Engine

from utilities.hypothesis import draw_and_map
from utilities.hypothesis.tempfile import temp_paths
from utilities.sqlalchemy import create_engine


@beartype
def sqlite_engines(
    *, metadata: MetaData | None = None, base: Any = None
) -> SearchStrategy[Engine]:
    """Strategy for generating SQLite engines."""

    return draw_and_map(
        _draw_sqlite_engines, temp_paths(), metadata=metadata, base=base
    )


@beartype
def _draw_sqlite_engines(
    temp_path: Path, /, *, metadata: MetaData | None = None, base: Any = None
) -> Engine:
    path = temp_path.joinpath("db.sqlite")
    engine = create_engine("sqlite", database=path.as_posix())
    if metadata is not None:
        metadata.create_all(engine)
    if base is not None:
        base.metadata.create_all(engine)

    # attach temp_path to the engine, so as to keep it alive
    engine.temp_path = temp_path  # type: ignore

    return engine



from re import compile

from beartype import beartype


@beartype
def extract_group(pattern: str, text: str, /) -> str:
    """Apply a regex with 1 capture group and check that there is exactly 1
    match, and then return it.
    """

    if compile(pattern).groups <= 1:
        (result,) = extract_groups(pattern, text)
        return result
    else:
        raise MultipleCaptureGroups(pattern)


class MultipleCaptureGroups(ValueError):
    ...


@beartype
def extract_groups(pattern: str, text: str, /) -> list[str]:
    """Apply a regex with a positive number of capture groups and check that
    there is exactly 1 match, and then return their contents as a single list.
    """

    compiled = compile(pattern)
    if (n_groups := compiled.groups) == 0:
        raise NoCaptureGroups(pattern)
    else:
        results = compiled.findall(text)
        if (n_results := len(results)) == 0:
            raise NoMatches(f"{pattern=}, {text=}")
        elif n_results == 1:
            if n_groups == 1:
                return results
            else:
                (result,) = results
                return list(result)
        else:
            raise MultipleMatches(f"{pattern=}, {text=}")


class NoCaptureGroups(ValueError):
    ...


class NoMatches(ValueError):
    ...


class MultipleMatches(ValueError):
    ...



import datetime as dt
from contextlib import suppress

from beartype import beartype

from utilities.re import extract_groups


@beartype
def ensure_date(date: dt.date | str, /) -> dt.date:
    """Ensure the object is a date."""

    return date if isinstance(date, dt.date) else parse_date(date)


@beartype
def ensure_datetime(datetime: dt.datetime | str, /) -> dt.datetime:
    """Ensure the object is a datetime."""

    if isinstance(datetime, dt.datetime):
        return datetime
    else:
        return parse_datetime(datetime)


@beartype
def ensure_time(time: dt.time | str, /) -> dt.time:
    """Ensure the object is a time."""

    return time if isinstance(time, dt.time) else parse_time(time)


@beartype
def ensure_timedelta(timedelta: dt.timedelta | str, /) -> dt.timedelta:
    """Ensure the object is a timedelta."""

    if isinstance(timedelta, dt.timedelta):
        return timedelta
    else:
        return parse_timedelta(timedelta)


@beartype
def parse_date(date: str, /) -> dt.date:
    """Parse a string into a date."""

    with suppress(ValueError):
        return dt.date.fromisoformat(date)
    with suppress(ValueError):
        return dt.datetime.strptime(date, "%Y%m%d").date()
    raise InvalidDate(date)


class InvalidDate(ValueError):
    ...


@beartype
def parse_datetime(datetime: str, /) -> dt.datetime:
    """Parse a string into a datetime."""

    with suppress(ValueError):
        return dt.datetime.fromisoformat(datetime)
    for format in [
        "%Y%m%d",
        "%Y%m%dT%H",
        "%Y%m%dT%H%M",
        "%Y%m%dT%H%M%S",
        "%Y%m%dT%H%M%S.%f",
    ]:
        with suppress(ValueError):
            return dt.datetime.strptime(datetime, format)
    raise InvalidDateTime(datetime)


class InvalidDateTime(ValueError):
    ...


@beartype
def parse_time(time: str) -> dt.time:
    """Parse a string into a time."""

    with suppress(ValueError):
        return dt.time.fromisoformat(time)
    for format in ["%H", "%H%M", "%H%M%S", "%H%M%S.%f"]:
        with suppress(ValueError):
            return dt.datetime.strptime(time, format).time()
    raise InvalidTime(time)


class InvalidTime(ValueError):
    ...


@beartype
def parse_timedelta(timedelta: str) -> dt.timedelta:
    """Parse a string into a timedelta."""

    for format in ["%H:%M:%S", "%H:%M:%S.%f"]:
        try:
            as_dt = dt.datetime.strptime(timedelta, format)
        except ValueError:
            pass
        else:
            return dt.timedelta(
                hours=as_dt.hour,
                minutes=as_dt.minute,
                seconds=as_dt.second,
                microseconds=as_dt.microsecond,
            )
    try:
        days, tail = extract_groups(
            r"([-\d]+)\s*(?:days?)?,?\s*([\d:\.]+)", timedelta
        )
    except ValueError:
        raise InvalidTimedelta(timedelta)
    else:
        return dt.timedelta(days=int(days)) + parse_timedelta(tail)


class InvalidTimedelta(ValueError):
    ...


@beartype
def serialize_date(date: dt.date, /) -> str:
    """Serialize a date."""

    return date.isoformat()


@beartype
def serialize_datetime(datetime: dt.datetime, /) -> str:
    """Serialize a datetime."""

    return datetime.isoformat()


@beartype
def serialize_time(time: dt.time, /) -> str:
    """Serialize a time."""

    return time.isoformat()


@beartype
def serialize_timedelta(timedelta: dt.timedelta, /) -> str:
    """Serialize a timedelta."""

    if (days := timedelta.days) != 0:
        tail = serialize_timedelta(timedelta - dt.timedelta(days=days))
        return f"d{days},{tail}"
    else:
        return str(timedelta)



from collections.abc import Iterator
from contextlib import contextmanager
from pathlib import Path

from atomicwrites import move_atomic
from atomicwrites import replace_atomic
from beartype import beartype

from utilities.pathlib import PathLike
from utilities.tempfile import TemporaryDirectory


@contextmanager
@beartype
def writer(path: PathLike, /, *, overwrite: bool = False) -> Iterator[Path]:
    """Yield a path for atomically writing files to disk."""

    path = Path(path)
    parent = path.parent
    parent.mkdir(parents=True, exist_ok=True)
    name = path.name
    with TemporaryDirectory(suffix=".tmp", prefix=name, dir=parent) as temp_dir:
        yield (temp_path := temp_dir.joinpath(name))
        src, dest = temp_path.as_posix(), path.as_posix()
        if overwrite:
            replace_atomic(src, dest)
        else:
            move_atomic(src, dest)



from enum import Enum
from typing import Any

from beartype import beartype


class StrEnum(str, Enum):
    """An enum whose elements are themselves strings."""

    @staticmethod
    @beartype
    def _generate_next_value_(
        name: str, start: Any, count: int, last_values: Any
    ) -> str:
        _ = start, count, last_values
        return name



from os import getenv

from hypothesis import Verbosity
from hypothesis import settings


kwargs = {"deadline": None, "print_blob": True, "report_multiple_bugs": False}
settings.register_profile("default", max_examples=100, **kwargs)
settings.register_profile("dev", max_examples=10, **kwargs)
settings.register_profile("ci", max_examples=1000, **kwargs)
settings.register_profile(
    "debug", max_examples=10, verbosity=Verbosity.verbose, **kwargs
)
settings.load_profile(getenv("HYPOTHESIS_PROFILE", "default"))



from collections.abc import Iterable
from collections.abc import Mapping
from dataclasses import dataclass
from io import StringIO
from io import TextIOWrapper
from typing import Any
from typing import cast

from beartype import beartype
from tqdm import tqdm as _tqdm

from utilities.pytest import is_pytest


@dataclass
@beartype
class _Defaults:
    desc: str | None = None
    total: int | float | None = None
    leave: bool | None = True
    file: TextIOWrapper | StringIO | None = None
    ncols: int | None = None
    mininterval: float | None = 0.1
    maxinterval: float | None = 10.0
    miniters: int | float | None = None
    ascii: bool | str | None = None
    unit: str | None = "it"
    unit_scale: bool | int | str | None = False
    dynamic_ncols: bool | None = False
    smoothing: float | None = 0.3
    bar_format: str | None = None
    initial: int | float | None = 0
    position: int | None = None
    postfix: Mapping[str, Any] | None = None
    unit_divisor: float | None = 1000.0
    write_bytes: bool | None = None
    lock_args: tuple[Any, ...] | None = None
    nrows: int | None = None
    colour: str | None = None
    delay: float | None = 0.0
    gui: bool | None = False


_DEFAULTS = _Defaults()


class tqdm(_tqdm):
    """Sub-class of `tqdm` which is disabled during pytest."""

    @beartype
    def __init__(
        self,
        iterable: Iterable[Any] | None = None,
        desc: str | None = _DEFAULTS.desc,
        total: int | float | None = _DEFAULTS.total,
        leave: bool | None = _DEFAULTS.leave,
        file: TextIOWrapper | StringIO | None = _DEFAULTS.file,
        ncols: int | None = _DEFAULTS.ncols,
        mininterval: float | None = _DEFAULTS.mininterval,
        maxinterval: float | None = _DEFAULTS.maxinterval,
        miniters: int | float | None = _DEFAULTS.miniters,
        ascii: bool | str | None = _DEFAULTS.ascii,
        unit: str | None = _DEFAULTS.unit,
        unit_scale: bool | int | str | None = _DEFAULTS.unit_scale,
        dynamic_ncols: bool | None = _DEFAULTS.dynamic_ncols,
        smoothing: float | None = _DEFAULTS.smoothing,
        bar_format: str | None = _DEFAULTS.bar_format,
        initial: int | float | None = 0,
        position: int | None = _DEFAULTS.position,
        postfix: Mapping[str, Any] | None = _DEFAULTS.postfix,
        unit_divisor: float | None = _DEFAULTS.unit_divisor,
        write_bytes: bool | None = _DEFAULTS.write_bytes,
        lock_args: tuple[Any, ...] | None = _DEFAULTS.lock_args,
        nrows: int | None = _DEFAULTS.nrows,
        colour: str | None = _DEFAULTS.colour,
        delay: float | None = _DEFAULTS.delay,
        gui: bool | None = _DEFAULTS.gui,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            iterable=cast(Any, iterable),
            desc=desc,
            total=total,
            leave=leave,
            file=file,
            ncols=ncols,
            mininterval=cast(Any, mininterval),
            maxinterval=cast(Any, maxinterval),
            miniters=miniters,
            ascii=ascii,
            disable=is_pytest(),
            unit=cast(Any, unit),
            unit_scale=cast(Any, unit_scale),
            dynamic_ncols=cast(Any, dynamic_ncols),
            smoothing=cast(Any, smoothing),
            bar_format=bar_format,
            initial=cast(Any, initial),
            position=position,
            postfix=postfix,
            unit_divisor=cast(Any, unit_divisor),
            write_bytes=write_bytes,
            lock_args=lock_args,
            nrows=nrows,
            colour=colour,
            delay=delay,
            gui=cast(Any, gui),
            **kwargs,
        )



from collections.abc import Iterator
from contextlib import contextmanager
from pathlib import Path
from zipfile import ZipFile

from beartype import beartype

from utilities.pathlib import PathLike
from utilities.tempfile import TemporaryDirectory


@contextmanager
@beartype
def yield_zip_file_contents(path: PathLike, /) -> Iterator[list[Path]]:
    """Yield the contents of a zipfile in a temporary directory."""

    with ZipFile(path) as zf, TemporaryDirectory() as temp:
        zf.extractall(path=temp)
        yield list(temp.iterdir())
    _ = zf  # make coverage understand this is returned



import datetime as dt
from collections.abc import Callable
from typing import Any

from hypothesis import given
from hypothesis.strategies import DataObject
from hypothesis.strategies import SearchStrategy
from hypothesis.strategies import data
from hypothesis.strategies import dates
from hypothesis.strategies import datetimes
from hypothesis.strategies import sampled_from
from hypothesis.strategies import timedeltas
from hypothesis.strategies import times
from pytest import mark
from pytest import param
from pytest import raises

from utilities.datetime import InvalidDate
from utilities.datetime import InvalidDateTime
from utilities.datetime import InvalidTime
from utilities.datetime import InvalidTimedelta
from utilities.datetime import ensure_date
from utilities.datetime import ensure_datetime
from utilities.datetime import ensure_time
from utilities.datetime import ensure_timedelta
from utilities.datetime import parse_date
from utilities.datetime import parse_datetime
from utilities.datetime import parse_time
from utilities.datetime import parse_timedelta
from utilities.datetime import serialize_date
from utilities.datetime import serialize_datetime
from utilities.datetime import serialize_time
from utilities.datetime import serialize_timedelta


class TestEnsure:
    @given(data=data())
    @mark.parametrize(
        ["strategy", "func"],
        [
            param(dates(), ensure_date),
            param(datetimes(), ensure_datetime),
            param(times(), ensure_time),
            param(timedeltas(), ensure_timedelta),
        ],
    )
    def test_main(
        self,
        data: DataObject,
        strategy: SearchStrategy[Any],
        func: Callable[[Any], Any],
    ) -> None:
        value = data.draw(strategy)
        input = data.draw(sampled_from([value, str(value)]))
        result = func(input)
        assert result == value


class TestParseDate:
    @given(date=dates())
    def test_str(self, date: dt.date) -> None:
        result = parse_date(str(date))
        assert result == date

    @given(date=dates())
    def test_isoformat(self, date: dt.date) -> None:
        result = parse_date(date.isoformat())
        assert result == date

    @given(date=dates())
    def test_yyyymmdd(self, date: dt.date) -> None:
        result = parse_date(date.strftime("%4Y%m%d"))
        assert result == date

    def test_error(self) -> None:
        with raises(InvalidDate):
            _ = parse_date("error")


class TestParseDateTime:
    @given(datetime=datetimes())
    def test_str(self, datetime: dt.datetime) -> None:
        result = parse_datetime(str(datetime))
        assert result == datetime

    @given(datetime=datetimes())
    def test_isoformat(self, datetime: dt.datetime) -> None:
        result = parse_datetime(datetime.isoformat())
        assert result == datetime

    @given(
        datetime=datetimes(),
        format=sampled_from(["%4Y%m%dT%H%M%S.%f", "%4Y-%m-%d %H:%M:%S.%f"]),
    )
    def test_yyyymmdd_hhmmss_fff(
        self, datetime: dt.datetime, format: str
    ) -> None:
        result = parse_datetime(datetime.strftime(format))
        assert result == datetime

    @given(
        datetime=datetimes(),
        format=sampled_from(
            ["%4Y%m%dT%H%M%S", "%4Y-%m-%d %H:%M:%S", "%4Y-%m-%dT%H:%M:%S"]
        ),
    )
    def test_yyyymmdd_hhmmss(self, datetime: dt.datetime, format: str) -> None:
        datetime = datetime.replace(microsecond=0)
        result = parse_datetime(datetime.strftime(format))
        assert result == datetime

    @given(
        datetime=datetimes(),
        format=sampled_from(
            ["%4Y%m%dT%H%M", "%4Y-%m-%d %H:%M", "%4Y-%m-%dT%H:%M"]
        ),
    )
    def test_yyyymmdd_hhmm(self, datetime: dt.datetime, format: str) -> None:
        datetime = datetime.replace(second=0, microsecond=0)
        result = parse_datetime(datetime.strftime(format))
        assert result == datetime

    @given(
        datetime=datetimes(),
        format=sampled_from(["%4Y%m%dT%H", "%4Y-%m-%d %H", "%4Y-%m-%dT%H"]),
    )
    def test_yyyymmdd_hh(self, datetime: dt.datetime, format: str) -> None:
        datetime = datetime.replace(minute=0, second=0, microsecond=0)
        result = parse_datetime(datetime.strftime(format))
        assert result == datetime

    @given(datetime=datetimes(), format=sampled_from(["%4Y%m%d", "%4Y-%m-%d"]))
    def test_yyyymmdd(self, datetime: dt.datetime, format: str) -> None:
        datetime = datetime.replace(hour=0, minute=0, second=0, microsecond=0)
        result = parse_datetime(datetime.strftime(format))
        assert result == datetime

    def test_error(self) -> None:
        with raises(InvalidDateTime):
            _ = parse_datetime("error")


class TestParseTime:
    @given(time=times())
    def test_str(self, time: dt.time) -> None:
        result = parse_time(str(time))
        assert result == time

    @given(time=times())
    def test_isoformat(self, time: dt.time) -> None:
        result = parse_time(time.isoformat())
        assert result == time

    @given(time=times(), format=sampled_from(["%H%M%S.%f", "%H:%M:%S.%f"]))
    def test_hhmmss_fff(self, time: dt.time, format: str) -> None:
        result = parse_time(time.strftime(format))
        assert result == time

    @given(time=times(), format=sampled_from(["%H%M%S", "%H:%M:%S"]))
    def test_hhmmss(self, time: dt.time, format: str) -> None:
        time = time.replace(microsecond=0)
        result = parse_time(time.strftime(format))
        assert result == time

    @given(time=times(), format=sampled_from(["%H%M", "%H:%M"]))
    def test_hhmm(self, time: dt.time, format: str) -> None:
        time = time.replace(second=0, microsecond=0)
        result = parse_time(time.strftime(format))
        assert result == time

    @given(time=times(), format=sampled_from(["%H", "%H"]))
    def test_hh(self, time: dt.time, format: str) -> None:
        time = time.replace(minute=0, second=0, microsecond=0)
        result = parse_time(time.strftime(format))
        assert result == time

    def test_error(self) -> None:
        with raises(InvalidTime):
            _ = parse_time("error")


class TestParseTimedelta:
    @given(timedelta=timedeltas())
    def test_main(self, timedelta: dt.timedelta) -> None:
        result = parse_timedelta(str(timedelta))
        assert result == timedelta

    def test_error(self) -> None:
        with raises(InvalidTimedelta):
            _ = parse_timedelta("error")


class TestSerialize:
    @given(data=data())
    @mark.parametrize(
        ["strategy", "serialize", "parse"],
        [
            param(dates(), serialize_date, parse_date),
            param(datetimes(), serialize_datetime, parse_datetime),
            param(times(), serialize_time, parse_time),
            param(timedeltas(), str, parse_timedelta),
            param(timedeltas(), serialize_timedelta, parse_timedelta),
        ],
    )
    def test_main(
        self,
        data: DataObject,
        strategy: SearchStrategy[Any],
        serialize: Callable[[Any], Any],
        parse: Callable[[Any], Any],
    ) -> None:
        value = data.draw(strategy)
        result = parse(serialize(value))
        assert result == value



import datetime as dt
from collections.abc import Iterator
from contextlib import contextmanager
from pathlib import Path

from beartype import beartype
from pyinstrument.profiler import Profiler

from utilities.pathlib import PathLike


@contextmanager
@beartype
def profile(*, path: PathLike = Path.cwd()) -> Iterator[None]:
    """Profile the contents of a block."""

    with Profiler() as profiler:
        yield
    now = dt.datetime.now()
    filename = Path(path, f"profile__{now:%Y%m%dT%H%M%S}.html")
    with open(filename, mode="w") as fh:
        _ = fh.write(profiler.output_html())
    _ = fh  # this ensures that `fh` is considered returned; for coverage



import datetime as dt

from beartype import beartype
from hypothesis.strategies import SearchStrategy
from hypothesis.strategies import dates
from hypothesis.strategies import datetimes

from utilities.pandas import TIMESTAMP_MAX_AS_DATE
from utilities.pandas import TIMESTAMP_MAX_AS_DATETIME
from utilities.pandas import TIMESTAMP_MIN_AS_DATE
from utilities.pandas import TIMESTAMP_MIN_AS_DATETIME


@beartype
def dates_pd(
    *,
    min_value: dt.date = TIMESTAMP_MIN_AS_DATE,
    max_value: dt.date = TIMESTAMP_MAX_AS_DATE,
) -> SearchStrategy[dt.date]:
    """Strategy for generating dates which can become Timestamps."""

    return dates(min_value=min_value, max_value=max_value)


@beartype
def datetimes_pd(
    *,
    min_value: dt.datetime = TIMESTAMP_MIN_AS_DATETIME,
    max_value: dt.datetime = TIMESTAMP_MAX_AS_DATETIME,
) -> SearchStrategy[dt.datetime]:
    """Strategy for generating datetimes which can become Timestamps."""

    return datetimes(min_value=min_value, max_value=max_value)



from collections.abc import Mapping
from pathlib import Path
from subprocess import PIPE  # noqa: S404
from subprocess import check_output  # noqa: S404

from beartype import beartype

from utilities.os import temp_environ
from utilities.pathlib import PathLike


@beartype
def get_shell_output(
    cmd: str,
    /,
    *,
    cwd: PathLike = Path.cwd(),
    activate: PathLike | None = None,
    env: Mapping[str, str | None] | None = None,
) -> str:
    """Get the output of a shell call. Activate a virtual environment if
    necessary.
    """

    cwd = Path(cwd)
    if activate is not None:
        activates = list(cwd.rglob("activate"))
        if (n := len(activates)) == 0:
            raise NoActivate(cwd)
        elif n == 1:
            cmd = f"source {activates[0]}; {cmd}"
        else:
            raise MultipleActivate(activates)
    with temp_environ(env):
        return check_output(
            cmd, stderr=PIPE, shell=True, cwd=cwd, text=True  # noqa: S602
        )


class NoActivate(ValueError):
    ...


class MultipleActivate(ValueError):
    ...



from beartype.door import die_if_unbearable
from beartype.roar import BeartypeAbbyHintViolation
from pytest import mark
from pytest import param
from pytest import raises

from utilities.types import Number


class TestNumber:
    @mark.parametrize("x", [param(0), param(0.0)])
    def test_success(self, x: Number) -> None:
        die_if_unbearable(x, Number)

    def test_error(self) -> None:
        with raises(BeartypeAbbyHintViolation):
            die_if_unbearable("0", Number)



import datetime as dt
from collections.abc import Callable
from collections.abc import Iterable
from itertools import starmap
from operator import itemgetter
from typing import Any
from typing import TypeVar
from typing import cast

from attr import NOTHING
from beartype import beartype
from cattrs import BaseConverter
from cattrs import Converter
from click import ParamType
from typed_settings import default_converter
from typed_settings import default_loaders
from typed_settings import load_settings as _load_settings
from typed_settings.click_utils import DEFAULT_TYPES
from typed_settings.click_utils import TypeHandler
from typed_settings.click_utils import click_options as _click_options

from utilities.click import Date
from utilities.click import DateTime
from utilities.click import Time
from utilities.click import Timedelta
from utilities.datetime import ensure_date
from utilities.datetime import ensure_datetime
from utilities.datetime import ensure_time
from utilities.datetime import ensure_timedelta
from utilities.datetime import serialize_date
from utilities.datetime import serialize_datetime
from utilities.datetime import serialize_time
from utilities.pathlib import PathLike


_T = TypeVar("_T")


@beartype
def load_settings(
    cls: type[_T], appname: str, /, *, config_files: Iterable[PathLike] = ()
) -> _T:
    """Load a settings object with the extended converter."""

    loaders = default_loaders(appname, config_files=config_files)
    converter = _make_converter()
    return _load_settings(cls, loaders, converter=converter)


@beartype
def click_options(
    cls: type[Any],
    appname: str,
    /,
    *,
    config_files: Iterable[PathLike] = (),
    argname: str | None = None,
) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
    """Generate click options with the extended converter."""

    loaders = default_loaders(appname, config_files=config_files)
    converter = _make_converter()
    type_handler = _make_type_handler()
    return _click_options(
        cls,
        loaders,
        converter=converter,
        type_handler=type_handler,
        argname=argname,
    )


@beartype
def _make_converter() -> BaseConverter | Converter:
    """Extend the default converter."""

    converter = default_converter()
    cases = [
        (dt.date, ensure_date),
        (dt.datetime, ensure_datetime),
        (dt.time, ensure_time),
        (dt.timedelta, ensure_timedelta),
    ]
    for cls, func in cases:
        hook = _make_structure_hook_for_class(cls, func)
        converter.register_structure_hook(cls, hook)
    return converter


@beartype
def _make_structure_hook_for_class(
    cls: type[Any], func: Callable[[Any], Any], /
) -> Callable[[Any, type[Any]], Any]:
    @beartype
    def hook(value: Any, _: type[Any] = Any, /) -> Any:
        if isinstance(value, (cls, str)):
            return func(value)
        else:
            raise TypeError(type(value))

    return hook


@beartype
def _make_type_handler() -> TypeHandler:
    cases = [
        (dt.date, Date, serialize_date),
        (dt.datetime, DateTime, serialize_datetime),
        (dt.time, Time, serialize_time),
        (dt.timedelta, Timedelta, str),
    ]
    extended = dict(
        zip(
            map(itemgetter(0), cases),
            starmap(_make_type_handler_for_class, cases),
        )
    )
    return TypeHandler(types=DEFAULT_TYPES | extended)


@beartype
def _make_type_handler_for_class(
    cls: type[Any], param: type[ParamType], serialize: Callable[[Any], str], /
) -> Callable[[Any, Any], dict[str, Any]]:
    @beartype
    def handler(_: type[Any], default: Any, /) -> dict[str, Any]:
        mapping = cast(dict[str, Any], {"type": param()})
        if default is not NOTHING and isinstance(  # pragma: no cover
            default, cls
        ):
            mapping["default"] = serialize(default)
        return mapping

    return handler



from pathlib import Path

from utilities.tempfile import TemporaryDirectory
from utilities.tempfile import gettempdir


class TestGetTempDir:
    def test_main(self) -> None:
        assert isinstance(gettempdir(), Path)


class TestTemporaryDirectory:
    def test_name(self) -> None:
        temp_dir = TemporaryDirectory()
        name = temp_dir.name
        assert isinstance(name, Path)
        assert name.is_dir()
        assert set(name.iterdir()) == set()

    def test_as_context_manager(self) -> None:
        with TemporaryDirectory() as temp:
            assert isinstance(temp, Path)
            assert temp.is_dir()
            assert set(temp.iterdir()) == set()
        assert not temp.is_dir()



from pathlib import Path

from pytest import raises

from utilities.subprocess import MultipleActivate
from utilities.subprocess import NoActivate
from utilities.subprocess import get_shell_output


class TestGetShellOutput:
    def test_main(self) -> None:
        output = get_shell_output("ls")
        assert any(line == "pyproject.toml" for line in output.splitlines())

    def test_activate(self, tmp_path: Path) -> None:
        venv = tmp_path.joinpath(".venv")
        activate = venv.joinpath("activate")
        activate.parent.mkdir(parents=True)
        activate.touch()
        _ = get_shell_output("ls", cwd=venv, activate=venv)

    def test_no_activate(self, tmp_path: Path) -> None:
        venv = tmp_path.joinpath(".venv")
        with raises(NoActivate):
            _ = get_shell_output("ls", cwd=venv, activate=venv)

    def test_multiple_activates(self, tmp_path: Path) -> None:
        venv = tmp_path.joinpath(".venv")
        for i in range(2):
            activate = venv.joinpath(str(i), "activate")
            activate.parent.mkdir(parents=True)
            activate.touch()
        with raises(MultipleActivate):
            _ = get_shell_output("ls", cwd=venv, activate=venv)



import builtins
from collections.abc import Callable
from collections.abc import Iterator
from contextlib import contextmanager
from functools import partial
from os import getenv
from re import search
from string import ascii_letters
from typing import Any
from typing import TypeVar

from beartype import beartype
from hypothesis import Verbosity
from hypothesis import assume
from hypothesis import settings
from hypothesis.strategies import SearchStrategy
from hypothesis.strategies import characters
from hypothesis.strategies import fixed_dictionaries
from hypothesis.strategies import just
from hypothesis.strategies import lists
from hypothesis.strategies import text
from hypothesis.strategies import tuples

from utilities.text import ensure_str


@contextmanager
@beartype
def assume_does_not_raise(
    *exceptions: type[Exception], match: str | None = None
) -> Iterator[None]:
    """Assume a set of exceptions are not raised. Optionally filter on the
    string representation of the exception.
    """

    try:
        yield
    except exceptions as caught:
        if match is None:
            _ = assume(False)
        else:
            (msg,) = caught.args
            if search(match, ensure_str(msg)):
                _ = assume(False)
            else:
                raise


_TD = TypeVar("_TD")


@beartype
def draw_and_map(
    func: Callable[..., _TD], *args: Any, **kwargs: Any
) -> SearchStrategy[_TD]:
    """Draw the elements and then pass them into a transformer."""

    return _lift_args_and_kwargs(*args, **kwargs).map(partial(_apply, func))


@beartype
def draw_and_flatmap(
    func: Callable[..., SearchStrategy[_TD]], *args: Any, **kwargs: Any
) -> SearchStrategy[_TD]:
    """Draw the elements and then flatmap them into a transformer."""

    return _lift_args_and_kwargs(*args, **kwargs).flatmap(partial(_apply, func))


@beartype
def _lift_args_and_kwargs(
    *args: Any, **kwargs: Any
) -> SearchStrategy[tuple[tuple[Any, ...], dict[str, Any]]]:
    lifted_args = tuples(*map(_lift, args))
    lifted_kwargs = fixed_dictionaries({k: _lift(v) for k, v in kwargs.items()})
    return tuples(lifted_args, lifted_kwargs)


@beartype
def _lift(x: Any, /) -> SearchStrategy[Any]:
    return x if isinstance(x, SearchStrategy) else just(x)


@beartype
def _apply(
    func: Callable[..., _TD], x: tuple[tuple[Any, ...], dict[str, Any]], /
) -> _TD:
    args, kwargs = x
    return func(*args, **kwargs)


_TLFL = TypeVar("_TLFL")


@beartype
def lists_fixed_length(
    strategy: SearchStrategy[_TLFL],
    size: int,
    /,
    *,
    unique: bool = False,
    sorted: bool = False,
) -> SearchStrategy[list[_TLFL]]:
    """Strategy for generating lists of a fixed length."""

    return draw_and_map(
        _draw_lists_fixed_length,
        lists(strategy, min_size=size, max_size=size, unique=unique),
        sorted=sorted,
    )


@beartype
def _draw_lists_fixed_length(
    elements: list[Any], /, *, sorted: bool = False
) -> list[Any]:
    if sorted:
        return builtins.sorted(elements)
    else:
        return elements


@beartype
def setup_hypothesis_profiles() -> None:
    """Set up the hypothesis profiles."""

    kwargs = {
        "deadline": None,
        "print_blob": True,
        "report_multiple_bugs": False,
    }
    settings.register_profile("default", max_examples=100, **kwargs)
    settings.register_profile("dev", max_examples=10, **kwargs)
    settings.register_profile("ci", max_examples=1000, **kwargs)
    settings.register_profile(
        "debug", max_examples=10, verbosity=Verbosity.verbose, **kwargs
    )
    settings.load_profile(getenv("HYPOTHESIS_PROFILE", "default"))


@beartype
def text_ascii(
    *, min_size: int = 0, max_size: int | None = None
) -> SearchStrategy[str]:
    """Strategy for generating ASCII text."""

    return text(
        characters(whitelist_categories=[], whitelist_characters=ascii_letters),
        min_size=min_size,
        max_size=max_size,
    )


@beartype
def text_clean(
    *, min_size: int = 0, max_size: int | None = None
) -> SearchStrategy[str]:
    """Strategy for generating clean text."""

    return text(
        characters(blacklist_categories=["Z", "C"]),
        min_size=min_size,
        max_size=max_size,
    )



from os import chdir
from pathlib import Path
from re import search
from time import sleep

from loguru import logger

from utilities.loguru import setup_loguru


class TestSetupLoguru:
    def test_main(self, tmp_path: Path) -> None:
        chdir(tmp_path)
        setup_loguru()

        sleep(1e-3)
        logger.debug("message")
        sleep(1e-3)

        (log,) = tmp_path.iterdir()
        assert log.name == "log"
        with open(log) as fh:
            (line,) = fh.read().splitlines()

        assert search(
            r"^\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\.\d{3} \| DEBUG\s* \| "
            + r"tests\.test_loguru:test_main:\d+ - message$",
            line,
        )



from pytest import mark
from pytest import param

from utilities.inflection import snake_case


class TestSnakeCase:
    @mark.parametrize(
        ["text", "expected"],
        [
            param("text", "text"),
            param("Text", "text"),
            param("text123", "text123"),
            param("Text123", "text123"),
            param("OneTwo", "one_two"),
            param("One Two", "one_two"),
            param("One  Two", "one_two"),
            param("One   Two", "one_two"),
            param("One_Two", "one_two"),
            param("One__Two", "one_two"),
            param("One___Two", "one_two"),
            param("HTML", "html"),
            param("NoHTML", "no_html"),
            param("HTMLVersion", "html_version"),
        ],
    )
    def test_main(self, text: str, expected: str) -> None:
        result = snake_case(text)
        assert result == expected



from collections.abc import Callable
from functools import reduce
from operator import ge
from operator import le
from re import search
from typing import Any

from sqlalchemy import Table
from sqlalchemy import and_
from sqlalchemy import case
from sqlalchemy import create_engine as _create_engine
from sqlalchemy.engine import URL
from sqlalchemy.engine import Engine
from sqlalchemy.exc import DatabaseError
from sqlalchemy.exc import OperationalError
from sqlalchemy.pool import NullPool
from sqlalchemy.pool import Pool


def columnwise_max(*columns: Any) -> Any:
    """Compute the columnwise max of a number of columns."""

    return _columnwise_minmax(*columns, op=ge)


def columnwise_min(*columns: Any) -> Any:
    """Compute the columnwise min of a number of columns."""

    return _columnwise_minmax(*columns, op=le)


def _columnwise_minmax(*columns: Any, op: Callable[[Any, Any], Any]) -> Any:
    """Compute the columnwise min of a number of columns."""

    def func(x: Any, y: Any, /) -> Any:
        x_none = x.is_(None)
        y_none = y.is_(None)
        col = case(
            (and_(x_none, y_none), None),
            (and_(~x_none, y_none), x),
            (and_(x_none, ~y_none), y),
            (op(x, y), x),
            else_=y,
        )
        # try auto-label
        names = {
            value
            for col in [x, y]
            if (value := getattr(col, "name", None)) is not None
        }
        try:
            (name,) = names
        except ValueError:
            return col
        else:
            return col.label(name)

    return reduce(func, columns)


def create_engine(
    drivername: str,
    /,
    *,
    username: str | None = None,
    password: str | None = None,
    host: str | None = None,
    port: int | None = None,
    database: str | None = None,
    poolclass: type[Pool] | None = NullPool,
) -> Engine:
    """Create a SQLAlchemy engine."""

    url = URL.create(
        drivername,
        username=username,
        password=password,
        host=host,
        port=port,
        database=database,
    )
    return _create_engine(url, future=True, poolclass=poolclass)


def ensure_table_created(table_or_model: Any, engine: Engine, /) -> None:
    """Ensure a table is created."""

    table = get_table(table_or_model)
    try:
        with engine.begin() as conn:
            table.create(conn)
    # note that OperationalError < DatabaseError
    except OperationalError as error:
        # sqlite
        (msg,) = error.args
        if not search("table .* already exists", msg):  # pragma: no cover
            raise
    except DatabaseError as error:  # pragma: no cover
        # oracle
        (msg,) = error.args
        if not search(
            "ORA-00955: name is already used by an existing object", msg
        ):
            raise


def ensure_table_dropped(table_or_model: Any, engine: Engine, /) -> None:
    """Ensure a table is dropped."""

    table = get_table(table_or_model)
    try:
        with engine.begin() as conn:
            table.drop(conn)
    # note that OperationalError < DatabaseError
    except OperationalError as error:
        # sqlite
        (msg,) = error.args
        if not search("no such table", msg):  # pragma: no cover
            raise
    except DatabaseError as error:  # pragma: no cover
        # oracle
        (msg,) = error.args
        if not search("ORA-00942: table or view does not exist", msg):
            raise


def get_column_names(table_or_model: Any, /) -> list[str]:
    """Get the column names from a table or model."""

    return [col.name for col in get_columns(table_or_model)]


def get_columns(table_or_model: Any, /) -> list[Any]:
    """Get the columns from a table or model."""

    return list(get_table(table_or_model).columns)


def get_table(table_or_model: Any, /) -> Table:
    """Get the table from a ORM model."""

    if isinstance(table_or_model, Table):
        return table_or_model
    else:
        return table_or_model.__table__



from enum import auto
from enum import unique

from utilities.enum import StrEnum


@unique
class LogLevel(StrEnum):
    """An enumeration of the logging levels."""

    DEBUG = auto()
    INFO = auto()
    WARNING = auto()
    ERROR = auto()
    CRITICAL = auto()



from beartype import beartype
from hypothesis.extra.numpy import array_shapes
from hypothesis.extra.numpy import arrays
from hypothesis.extra.numpy import from_dtype
from hypothesis.strategies import SearchStrategy
from hypothesis.strategies import booleans
from hypothesis.strategies import floats
from hypothesis.strategies import integers
from hypothesis.strategies import nothing
from numpy import bool_
from numpy import dtype
from numpy import float64
from numpy import iinfo
from numpy import int64
from numpy.typing import NDArray

from utilities.hypothesis import draw_and_flatmap
from utilities.hypothesis.typing import MaybeSearchStrategy
from utilities.hypothesis.typing import Shape


@beartype
def bool_arrays(
    *, shape: MaybeSearchStrategy[Shape] = array_shapes()
) -> SearchStrategy[NDArray[bool_]]:
    """Strategy for generating arrays of booleans."""

    return draw_and_flatmap(_draw_bool_arrays, shape)


@beartype
def _draw_bool_arrays(shape: Shape, /) -> SearchStrategy[NDArray[bool_]]:
    return arrays(bool, shape, elements=booleans(), fill=nothing())


@beartype
def float_arrays(
    *,
    shape: MaybeSearchStrategy[Shape] = array_shapes(),
    min_value: MaybeSearchStrategy[float | None] = None,
    max_value: MaybeSearchStrategy[float | None] = None,
    allow_nan: MaybeSearchStrategy[bool | None] = None,
    allow_infinity: MaybeSearchStrategy[bool | None] = None,
) -> SearchStrategy[NDArray[float64]]:
    """Strategy for generating arrays of floats."""

    return draw_and_flatmap(
        _draw_float_arrays,
        shape,
        min_value=min_value,
        max_value=max_value,
        allow_nan=allow_nan,
        allow_infinity=allow_infinity,
    )


@beartype
def _draw_float_arrays(
    shape: Shape,
    /,
    *,
    min_value: float | None = None,
    max_value: float | None = None,
    allow_nan: bool | None = None,
    allow_infinity: bool | None = None,
) -> SearchStrategy[NDArray[float64]]:
    elements = floats(
        min_value=min_value,
        max_value=max_value,
        allow_nan=allow_nan,
        allow_infinity=allow_infinity,
    )
    return arrays(float, shape, elements=elements, fill=nothing())


@beartype
def int_arrays(
    *,
    shape: MaybeSearchStrategy[Shape] = array_shapes(),
    min_value: MaybeSearchStrategy[int | None] = None,
    max_value: MaybeSearchStrategy[int | None] = None,
) -> SearchStrategy[NDArray[int64]]:
    """Strategy for generating arrays of ints."""

    return draw_and_flatmap(
        _draw_int_arrays, shape, min_value=min_value, max_value=max_value
    )


@beartype
def _draw_int_arrays(
    shape: Shape,
    /,
    *,
    min_value: int | None = None,
    max_value: int | None = None,
) -> SearchStrategy[NDArray[int64]]:
    info = iinfo(int64)
    min_value_use = info.min if min_value is None else min_value
    max_value_use = info.max if max_value is None else max_value
    elements = integers(min_value=min_value_use, max_value=max_value_use)
    return arrays(int, shape, elements=elements, fill=nothing())


@beartype
def int64s() -> SearchStrategy[int]:
    """Strategy for generating int64s."""

    return from_dtype(dtype(int64)).map(int)



from pathlib import Path
from uuid import UUID

from beartype import beartype
from hypothesis.strategies import SearchStrategy
from hypothesis.strategies import uuids

from utilities.hypothesis import draw_and_map
from utilities.tempfile import TemporaryDirectory
from utilities.tempfile import gettempdir


@beartype
def temp_dirs() -> SearchStrategy[TemporaryDirectory]:
    """Search strategy for temporary directories."""

    dir = gettempdir().joinpath("hypothesis")
    dir.mkdir(exist_ok=True)
    return draw_and_map(_draw_temp_dirs, uuids(), dir)


@beartype
def _draw_temp_dirs(uuid: UUID, dir: Path, /) -> TemporaryDirectory:
    return TemporaryDirectory(prefix=f"{uuid}__", dir=dir.as_posix())


@beartype
def temp_paths() -> SearchStrategy[Path]:
    """Search strategy for paths to temporary directories."""

    return draw_and_map(_draw_temp_paths, temp_dirs())


@beartype
def _draw_temp_paths(temp_dir: TemporaryDirectory, /) -> Path:
    class SubPath(type(root := temp_dir.name)):
        _temp_dir = temp_dir

    return SubPath(root)



import datetime as dt
from json import dumps
from pathlib import Path
from typing import Any

from beartype import beartype


@beartype
def serialize(
    x: Any,
    /,
    *,
    skipkeys: bool = False,
    ensure_ascii: bool = True,
    check_circular: bool = True,
    allow_nan: bool = True,
    indent: int | str | None = None,
    separators: tuple[str, str] | None = None,
    sort_keys: bool = False,
    **kwargs: Any,
) -> str:
    """Serialize an object."""

    return dumps(
        x,
        skipkeys=skipkeys,
        ensure_ascii=ensure_ascii,
        check_circular=check_circular,
        allow_nan=allow_nan,
        indent=indent,
        separators=separators,
        default=_default,
        sort_keys=sort_keys,
        **kwargs,
    )


@beartype
def _default(x: Any, /) -> str:
    """Extension for the JSON serializer."""

    if isinstance(x, dt.date):
        return x.isoformat()
    elif isinstance(x, Path):
        return x.as_posix()
    elif isinstance(x, set):
        inner = serialize(sorted(x))
        return f"set({inner})"
    elif isinstance(x, frozenset):
        inner = serialize(sorted(x))
        return f"frozenset({inner})"
    else:
        raise TypeError(f"Invalid type: {x}")



from os import getenv

from hypothesis import given

from utilities.hypothesis import text_ascii
from utilities.os import temp_environ


text = text_ascii(min_size=1, max_size=10)


def prefix(text: str, /) -> str:
    return f"_TEST_OS_{text}"


class TestTempEnviron:
    @given(key=text.map(prefix), value=text)
    def test_set(self, key: str, value: str) -> None:
        assert getenv(key) is None
        with temp_environ({key: value}):
            assert getenv(key) == value
        assert getenv(key) is None

    @given(key=text.map(prefix), prev=text, new=text)
    def test_override(self, key: str, prev: str, new: str) -> None:
        with temp_environ({key: prev}):
            assert getenv(key) == prev
            with temp_environ({key: new}):
                assert getenv(key) == new
            assert getenv(key) == prev

    @given(key=text.map(prefix), value=text)
    def test_unset(self, key: str, value: str) -> None:
        with temp_environ({key: value}):
            assert getenv(key) == value
            with temp_environ({key: None}):
                assert getenv(key) is None
            assert getenv(key) == value



from utilities.enum import StrEnum
from utilities.logging import LogLevel


class TestLogLevel:
    def test_main(self) -> None:
        assert issubclass(LogLevel, StrEnum)
        assert len(LogLevel) == 5



from enum import auto

from hypothesis import given
from hypothesis.strategies import sampled_from

from utilities.enum import StrEnum


class Truth(StrEnum):
    true = auto()
    false = auto()


class TestStrEnum:
    @given(truth=sampled_from(Truth))
    def test_main(self, truth: Truth) -> None:
        assert isinstance(truth, str)
        assert truth == truth.name



from pathlib import Path
from tempfile import TemporaryDirectory as _TemporaryDirectory
from tempfile import gettempdir as _gettempdir

from beartype import beartype

from utilities.pathlib import PathLike


class TemporaryDirectory(_TemporaryDirectory):
    """Sub-class of TemporaryDirectory whose name attribute is a Path."""

    name: Path

    @beartype
    def __init__(
        self,
        *,
        suffix: str | None = None,
        prefix: str | None = None,
        dir: PathLike | None = None,
        ignore_cleanup_errors: bool = False,
    ) -> None:
        super().__init__(
            suffix=suffix,
            prefix=prefix,
            dir=dir,
            ignore_cleanup_errors=ignore_cleanup_errors,
        )
        self.name = Path(self.name)

    @beartype
    def __enter__(self) -> Path:
        return super().__enter__()


@beartype
def gettempdir() -> Path:
    """Get the name of the directory used for temporary files."""

    return Path(_gettempdir())



import datetime as dt
from collections.abc import Callable
from pathlib import Path
from typing import Any
from typing import cast

from click import command
from click import echo
from click.testing import CliRunner
from hypothesis import given
from hypothesis.strategies import DataObject
from hypothesis.strategies import SearchStrategy
from hypothesis.strategies import data
from hypothesis.strategies import dates
from hypothesis.strategies import datetimes
from hypothesis.strategies import timedeltas
from hypothesis.strategies import times
from hypothesis.strategies import tuples
from pytest import mark
from pytest import param
from pytest import raises
from typed_settings import settings
from typed_settings.exceptions import InvalidValueError

from utilities.datetime import serialize_date
from utilities.datetime import serialize_datetime
from utilities.datetime import serialize_time
from utilities.datetime import serialize_timedelta
from utilities.hypothesis.tempfile import temp_paths
from utilities.typed_settings import click_options
from utilities.typed_settings import load_settings


class TestLoadSettings:
    @given(data=data(), root=temp_paths())
    @mark.parametrize(
        ["cls", "strategy"],
        [
            param(dt.date, dates()),
            param(dt.datetime, datetimes()),
            param(dt.time, times()),
            param(dt.timedelta, timedeltas()),
        ],
    )
    def test_main(
        self,
        data: DataObject,
        root: Path,
        cls: Any,
        strategy: SearchStrategy[Any],
    ) -> None:
        default, value = data.draw(tuples(strategy, strategy))

        @settings(frozen=True)
        class Settings:
            value: cls = default

        settings_default = load_settings(Settings, "appname")
        assert settings_default.value == default
        with open(file := root.joinpath("file.toml"), mode="w") as fh:
            _ = fh.write(f'[appname]\nvalue = "{value}"')
        settings_loaded = load_settings(
            Settings, "appname", config_files=[file]
        )
        assert settings_loaded.value == value

    @mark.parametrize(
        "cls", [param(dt.date), param(dt.time), param(dt.timedelta)]
    )
    def test_errors(self, cls: Any) -> None:
        @settings(frozen=True)
        class Settings:
            value: cls = cast(Any, None)

        with raises(InvalidValueError):
            _ = load_settings(Settings, "appname")


class TestClickOptions:
    @given(data=data(), root=temp_paths())
    @mark.parametrize(
        ["cls", "strategy", "serialize"],
        [
            param(dt.date, dates(), serialize_date),
            param(dt.datetime, datetimes(), serialize_datetime),
            param(dt.time, times(), serialize_time),
            param(dt.timedelta, timedeltas(), serialize_timedelta),
        ],
    )
    def test_main(
        self,
        data: DataObject,
        root: Path,
        cls: Any,
        strategy: SearchStrategy[Any],
        serialize: Callable[[Any], str],
    ) -> None:
        default, val, cfg = data.draw(tuples(strategy, strategy, strategy))
        val_str, cfg_str = map(serialize, [val, cfg])
        runner = CliRunner()

        @settings(frozen=True)
        class Settings:
            value: cls = default

        @command()
        @click_options(Settings, "appname")
        def cli1(settings: Settings, /) -> None:
            echo(f"value = {serialize(settings.value)}")

        result = runner.invoke(cli1)
        assert result.exit_code == 0
        assert result.stdout == f"value = {serialize(default)}\n"

        result = runner.invoke(cli1, f'--value="{val_str}"')
        assert result.exit_code == 0
        assert result.stdout == f"value = {val_str}\n"

        with open(file := root.joinpath("file.toml"), mode="w") as fh:
            _ = fh.write(f'[appname]\nvalue = "{cfg_str}"')

        @command()
        @click_options(Settings, "appname", config_files=[file])
        def cli2(settings: Settings, /) -> None:
            echo(f"value = {serialize(settings.value)}")

        result = runner.invoke(cli2)
        assert result.exit_code == 0
        assert result.stdout == f"value = {cfg_str}\n"

        result = runner.invoke(cli1, f'--value="{val_str}"')
        assert result.exit_code == 0
        assert result.stdout == f"value = {val_str}\n"



import builtins
from re import search

from hypothesis import given
from hypothesis.errors import InvalidArgument
from hypothesis.strategies import DataObject
from hypothesis.strategies import SearchStrategy
from hypothesis.strategies import booleans
from hypothesis.strategies import data
from hypothesis.strategies import integers
from hypothesis.strategies import just
from hypothesis.strategies import none
from pytest import mark
from pytest import param
from pytest import raises

from utilities.hypothesis import assume_does_not_raise
from utilities.hypothesis import draw_and_flatmap
from utilities.hypothesis import draw_and_map
from utilities.hypothesis import lists_fixed_length
from utilities.hypothesis import setup_hypothesis_profiles
from utilities.hypothesis import text_ascii
from utilities.hypothesis import text_clean


class TestAssumeDoesNotRaise:
    @given(x=booleans())
    def test_no_match_and_suppressed(self, x: bool) -> None:
        with assume_does_not_raise(ValueError):
            if x is True:
                raise ValueError("x is True")
        assert x is False

    @given(x=just(True))
    def test_no_match_and_not_suppressed(self, x: bool) -> None:
        with raises(ValueError, match="x is True"), assume_does_not_raise(
            RuntimeError
        ):
            if x is True:
                raise ValueError("x is True")

    @given(x=booleans())
    def test_with_match_and_suppressed(self, x: bool) -> None:
        with assume_does_not_raise(ValueError, match="x is True"):
            if x is True:
                raise ValueError("x is True")
        assert x is False

    @given(x=just(True))
    def test_with_match_and_not_suppressed(self, x: bool) -> None:
        with raises(ValueError, match="x is True"), assume_does_not_raise(
            ValueError, match="wrong"
        ):
            if x is True:
                raise ValueError("x is True")


def uses_draw_and_map(
    x: bool | SearchStrategy[bool], /
) -> SearchStrategy[bool]:
    def inner(x: bool, /) -> bool:
        return x

    return draw_and_map(inner, x)


class TestDrawAndMap:
    @given(data=data(), x=booleans())
    def test_fixed(self, data: DataObject, x: bool) -> None:
        result = data.draw(uses_draw_and_map(x))
        assert result is x

    @given(x=uses_draw_and_map(booleans()))
    def test_strategy(self, x: bool) -> None:
        assert isinstance(x, bool)


def uses_draw_and_flatmap(
    x: bool | SearchStrategy[bool], /
) -> SearchStrategy[bool]:
    def inner(x: bool, /) -> SearchStrategy[bool]:
        return just(x)

    return draw_and_flatmap(inner, x)


class TestDrawAndFlatMap:
    @given(data=data(), x=booleans())
    def test_fixed(self, data: DataObject, x: bool) -> None:
        result = data.draw(uses_draw_and_flatmap(x))
        assert result is x

    @given(x=uses_draw_and_flatmap(booleans()))
    def test_strategy(self, x: bool) -> None:
        assert isinstance(x, bool)


class TestListsFixedLength:
    @given(data=data(), size=integers(1, 10))
    @mark.parametrize(
        "unique", [param(True, id="unique"), param(False, id="no unique")]
    )
    @mark.parametrize(
        "sorted", [param(True, id="sorted"), param(False, id="no sorted")]
    )
    def test_main(
        self, data: DataObject, size: int, unique: bool, sorted: bool
    ) -> None:
        result = data.draw(
            lists_fixed_length(integers(), size, unique=unique, sorted=sorted)
        )
        assert isinstance(result, list)
        assert len(result) == size
        if unique:
            assert len(set(result)) == len(result)
        if sorted:
            assert builtins.sorted(result) == result


class TestSetupHypothesisProfiles:
    def test_main(self) -> None:
        setup_hypothesis_profiles()


class TestTextAscii:
    @given(
        data=data(),
        min_size=integers(0, 100),
        max_size=integers(0, 100) | none(),
    )
    def test_main(
        self, data: DataObject, min_size: int, max_size: int | None
    ) -> None:
        with assume_does_not_raise(InvalidArgument, AssertionError):
            text = data.draw(text_ascii(min_size=min_size, max_size=max_size))
        assert search("^[A-Za-z]*$", text)
        assert len(text) >= min_size
        if max_size is not None:
            assert len(text) <= max_size


class TestTextClean:
    @given(
        data=data(),
        min_size=integers(0, 100),
        max_size=integers(0, 100) | none(),
    )
    def test_main(
        self, data: DataObject, min_size: int, max_size: int | None
    ) -> None:
        with assume_does_not_raise(InvalidArgument, AssertionError):
            text = data.draw(text_clean(min_size=min_size, max_size=max_size))
        assert search("^\\S[^\\r\\n]*$|^$", text)
        assert len(text) >= min_size
        if max_size is not None:
            assert len(text) <= max_size



from pathlib import Path

from hypothesis import given
from hypothesis.strategies import sets

from utilities.hypothesis import text_ascii
from utilities.hypothesis.tempfile import temp_dirs
from utilities.hypothesis.tempfile import temp_paths
from utilities.tempfile import TemporaryDirectory


class TestTempDirs:
    @given(temp_dir=temp_dirs())
    def test_main(self, temp_dir: TemporaryDirectory) -> None:
        _test_temp_path(temp_dir.name)

    @given(
        temp_dir=temp_dirs(), contents=sets(text_ascii(min_size=1), max_size=10)
    )
    def test_writing_files(
        self, temp_dir: TemporaryDirectory, contents: set[str]
    ) -> None:
        _test_writing_to_temp_path(temp_dir.name, contents)


class TestTempPaths:
    @given(temp_path=temp_paths())
    def test_main(self, temp_path: Path) -> None:
        _test_temp_path(temp_path)

    @given(
        temp_path=temp_paths(),
        contents=sets(text_ascii(min_size=1), max_size=10),
    )
    def test_writing_files(self, temp_path: Path, contents: set[str]) -> None:
        _test_writing_to_temp_path(temp_path, contents)


def _test_temp_path(path: Path, /) -> None:
    assert path.is_dir()
    assert len(set(path.iterdir())) == 0


def _test_writing_to_temp_path(path: Path, contents: set[str], /) -> None:
    assert len(set(path.iterdir())) == 0
    for content in contents:
        path.joinpath(content).touch()
    assert len(set(path.iterdir())) == len(contents)



from textwrap import dedent
from typing import Any

from beartype import beartype


@beartype
def ensure_str(x: Any, /) -> str:
    """Ensure an object is a string."""

    if isinstance(x, str):
        return x
    else:
        raise TypeError(f"{x=}")


class NotAString(TypeError):
    ...


@beartype
def strip_and_dedent(text: str, /) -> str:
    """Strip and dedent a string."""

    return dedent(text.strip("\n")).strip("\n")



from typing import Any

from numpy import array
from pandas import DatetimeTZDtype
from pandas import Series
from pytest import mark
from pytest import param

from utilities.numpy import has_dtype


class TestHasDtype:
    @mark.parametrize(
        "x, dtype, expected",
        [
            param(array([]), float, True),
            param(array([]), (float,), True),
            param(array([]), int, False),
            param(array([]), (int,), False),
            param(array([]), "Int64", False),
            param(array([]), ("Int64",), False),
            param(Series([], dtype="Int64"), "Int64", True),
            param(Series([], dtype="Int64"), int, False),
            param(
                Series([], dtype=DatetimeTZDtype(tz="UTC")),
                DatetimeTZDtype(tz="UTC"),
                True,
            ),
            param(
                Series([], dtype=DatetimeTZDtype(tz="UTC")),
                DatetimeTZDtype(tz="Asia/Hong_Kong"),
                False,
            ),
        ],
    )
    def test_main(self, x: Any, dtype: Any, expected: bool) -> None:
        assert has_dtype(x, dtype) is expected



import datetime as dt
from typing import Any
from typing import cast

import sqlalchemy
from hypothesis import assume
from hypothesis import given
from hypothesis.extra.pandas import column
from hypothesis.extra.pandas import data_frames
from hypothesis.extra.pandas import range_indexes
from hypothesis.strategies import integers
from hypothesis.strategies import sets
from numpy import inf
from numpy import nan
from pandas import NA
from pandas import DataFrame
from pandas import Series
from pandas import to_datetime
from pytest import mark
from pytest import param
from pytest import raises
from sqlalchemy import Boolean
from sqlalchemy import Column
from sqlalchemy import Date
from sqlalchemy import DateTime
from sqlalchemy import Float
from sqlalchemy import Integer
from sqlalchemy import String
from sqlalchemy import select
from sqlalchemy.engine import Engine
from sqlalchemy.orm import Session
from sqlalchemy.orm import declarative_base

from utilities.hypothesis.sqlalchemy import sqlite_engines
from utilities.numpy import datetime64ns
from utilities.pandas import Int64
from utilities.pandas import boolean
from utilities.pandas import string
from utilities.sqlalchemy import get_table
from utilities.sqlalchemy.pandas import _get_dtype
from utilities.sqlalchemy.pandas import _nativize_column
from utilities.sqlalchemy.pandas import insert_dataframe
from utilities.sqlalchemy.pandas import read_dataframe


class TestGetDType:
    @mark.parametrize(
        "column, expected",
        [
            param(Column(Boolean), boolean),
            param(Column(Date), datetime64ns),
            param(Column(DateTime), datetime64ns),
            param(Column(Float), float),
            param(Column(Integer), Int64),
            param(Column(String), string),
            param(Column(sqlalchemy.DECIMAL), float),
        ],
    )
    def test_main(self, column: Any, expected: Any) -> None:
        assert _get_dtype(column) == expected


class TestInsertDataFrame:
    @given(
        df=data_frames(
            [column(name="id", dtype=int)],  # type: ignore
            index=range_indexes(max_size=10),
        ),
        engine=sqlite_engines(),
    )
    def test_main(self, df: DataFrame, engine: Engine) -> None:
        _ = assume(not df["id"].duplicated().any())

        class Example(cast(Any, declarative_base())):
            __tablename__ = "example"
            id = Column(Integer, primary_key=True)

        with engine.begin() as conn:
            get_table(Example).create(conn)

        insert_dataframe(df, Example, engine)

        with engine.begin() as conn:
            res = conn.execute(select(Example)).all()
        assert len(res) == len(df)


class TestNativizeColumn:
    @mark.parametrize(
        "series, expected",
        [
            param(Series([True, False], dtype=bool), [True, False]),
            param(
                Series([True, False, None], dtype=boolean), [True, False, None]
            ),
        ],
    )
    @mark.parametrize(
        "column", [param(Column(Boolean)), param(Column(Integer))]
    )
    def test_boolean_data(
        self, series: Series, column: Any, expected: list[Any]
    ) -> None:
        res = list(_nativize_column(series, column))
        assert res == expected

    @mark.parametrize(
        "series, column, expected",
        [
            param(
                Series([to_datetime("2000-01-01"), NA], dtype=datetime64ns),
                Column(Date),
                [dt.date(2000, 1, 1), None],
            ),
            param(
                Series(
                    [to_datetime("2000-01-01 12:00:00"), NA], dtype=datetime64ns
                ),
                Column(DateTime),
                [dt.datetime(2000, 1, 1, 12), None],
            ),
        ],
    )
    def test_datetime_data(
        self, series: Series, column: Any, expected: list[Any]
    ) -> None:
        res = list(_nativize_column(series, column))
        assert res == expected

    @mark.parametrize(
        "series, column, expected",
        [
            param(
                Series([0.0, nan, inf, -inf], dtype=float),
                Column(Float),
                [0.0, None, inf, -inf],
            ),
            param(Series([0], dtype=int), Column(Integer), [0]),
            param(Series([0, NA], dtype=Int64), Column(Integer), [0, None]),
            param(
                Series(["string", NA], dtype=string),
                Column(String),
                ["string", None],
            ),
        ],
    )
    def test_float_int_and_str_data(
        self, series: Series, column: Any, expected: list[Any]
    ) -> None:
        res = list(_nativize_column(series, column))
        assert res == expected

    def test_error(self) -> None:
        series = Series([True, False], dtype=bool)
        column = Column(String)
        with raises(TypeError, match="Invalid types: .*"):
            _ = list(_nativize_column(series, column))


class TestReadDataFrame:
    @given(ids=sets(integers(0, 100), max_size=10), engine=sqlite_engines())
    def test_main(self, ids: set[int], engine: Engine) -> None:
        class Example(cast(Any, declarative_base())):
            __tablename__ = "example"
            id = Column(Integer, primary_key=True)

        with engine.begin() as conn:
            get_table(Example).create(conn)
        with Session(engine) as sess:
            sess.add_all([Example(id=id) for id in ids])
            sess.commit()

        df = read_dataframe(select(Example), engine)
        assert len(df) == len(ids)
        assert dict(df.dtypes) == {"id": Int64}



from pytest import mark
from pytest import param
from pytest import raises

from utilities.re import MultipleCaptureGroups
from utilities.re import MultipleMatches
from utilities.re import NoCaptureGroups
from utilities.re import NoMatches
from utilities.re import extract_group
from utilities.re import extract_groups


class TestExtractGroup:
    def test_success(self) -> None:
        assert extract_group(r"(\d)", "A0A") == "0"

    def test_no_groups(self) -> None:
        with raises(NoCaptureGroups):
            _ = extract_group(r"\d", "0")

    def test_multiple_groups(self) -> None:
        with raises(MultipleCaptureGroups):
            _ = extract_group(r"(\d)(\w)", "0A")

    def test_no_matches(self) -> None:
        with raises(NoMatches, match="pattern='.*', text='.*'"):
            _ = extract_group(r"(\d)", "A")

    def test_multiple_matches(self) -> None:
        with raises(MultipleMatches, match="pattern='.*', text='.*'"):
            _ = extract_group(r"(\d)", "0A0")


class TestExtractGroups:
    @mark.parametrize(
        "pattern, text, expected",
        [param(r"(\d)", "A0A", ["0"]), param(r"(\d)(\w)", "A0A0", ["0", "A"])],
    )
    def test_success(
        self, pattern: str, text: str, expected: list[str]
    ) -> None:
        assert extract_groups(pattern, text) == expected

    def test_no_groups(self) -> None:
        with raises(NoCaptureGroups):
            _ = extract_groups(r"\d", "0")

    @mark.parametrize(
        "pattern, text", [param(r"(\d)", "A"), param(r"(\d)(\w)", "A0")]
    )
    def test_no_matches(self, pattern: str, text: str) -> None:
        with raises(NoMatches, match="pattern='.*', text='.*'"):
            _ = extract_groups(pattern, text)

    @mark.parametrize(
        "pattern, text", [param(r"(\d)", "0A0"), param(r"(\d)(\w)", "0A0A")]
    )
    def test_multiple_matches(self, pattern: str, text: str) -> None:
        with raises(MultipleMatches, match="pattern='.*', text='.*'"):
            _ = extract_groups(pattern, text)



import datetime as dt
from collections.abc import Callable
from operator import eq
from operator import ge
from operator import gt
from operator import le
from operator import lt
from operator import ne
from re import search
from time import sleep
from typing import Any

from pytest import mark
from pytest import param
from pytest import raises

from utilities.timer import Timer


class TestTimer:
    def test_context_manager(self) -> None:
        with Timer() as timer:
            assert isinstance(timer, Timer)
            sleep(1e-3)
        assert timer >= 1e-3

    @mark.parametrize(
        "op, expected",
        [
            param(eq, False),
            param(ne, True),
            param(ge, False),
            param(gt, False),
            param(le, True),
            param(lt, True),
        ],
    )
    @mark.parametrize(
        "dur", [param(1), param(1.0), param(dt.timedelta(seconds=1))]
    )
    def test_comparison(
        self, op: Callable[[Any, Any], bool], dur: Any, expected: bool
    ) -> None:
        with Timer() as timer:
            pass
        assert op(timer, dur) is expected

    def test_comparison_between_timers(self) -> None:
        with Timer() as timer1:
            pass
        with Timer() as timer2:
            pass
        assert isinstance(timer1 == timer2, bool)

    def test_comparison_error(self) -> None:
        with Timer() as timer:
            pass
        with raises(TypeError):
            _ = timer == "error"

    @mark.parametrize("func", [param(repr), param(str)])
    def test_repr_and_str(self, func: Callable[[Timer], str]) -> None:
        with Timer() as timer:
            pass
        as_str = func(timer)
        assert search(r"^\d+:\d{2}:\d{2}\.\d{6}$", as_str)

    def test_running(self) -> None:
        timer = Timer()
        sleep(1e-3)
        assert timer >= 1e-3
        sleep(1e-3)
        assert timer >= 2e-3



from pathlib import Path
from typing import Any
from typing import cast

from hypothesis import given
from hypothesis.strategies import fixed_dictionaries
from hypothesis.strategies import integers
from hypothesis.strategies import lists
from hypothesis.strategies import none
from pytest import mark
from pytest import param
from pytest import raises
from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy import MetaData
from sqlalchemy import Table
from sqlalchemy import select
from sqlalchemy.engine import Engine
from sqlalchemy.exc import OperationalError
from sqlalchemy.orm import declarative_base

from utilities.hypothesis.sqlalchemy import sqlite_engines
from utilities.hypothesis.tempfile import temp_paths
from utilities.sqlalchemy import columnwise_max
from utilities.sqlalchemy import columnwise_min
from utilities.sqlalchemy import create_engine
from utilities.sqlalchemy import ensure_table_created
from utilities.sqlalchemy import ensure_table_dropped
from utilities.sqlalchemy import get_column_names
from utilities.sqlalchemy import get_columns
from utilities.sqlalchemy import get_table


class TestColumnwiseMinMax:
    @given(
        values=lists(
            fixed_dictionaries(
                {"x": integers(0, 100) | none(), "y": integers(0, 100) | none()}
            ),
            min_size=1,
            max_size=10,
        ),
        engine=sqlite_engines(),
    )
    def test_main(
        self, values: list[dict[str, int | None]], engine: Engine
    ) -> None:
        table = Table(
            "example",
            MetaData(),
            Column("id", Integer, primary_key=True, autoincrement=True),
            Column("x", Integer),
            Column("y", Integer),
        )
        ensure_table_created(table, engine)
        with engine.begin() as conn:
            _ = conn.execute(table.insert(), values)
            res = conn.execute(table.select()).all()

        sel = select(
            table.c.x,
            table.c.y,
            columnwise_min(table.c.x, table.c.y).label("min_xy"),
            columnwise_max(table.c.x, table.c.y).label("max_xy"),
        )
        with engine.begin() as conn:
            res = conn.execute(sel).all()

        assert len(res) == len(values)
        for x, y, min_xy, max_xy in res:
            if (x is None) and (y is None):
                assert min_xy is None
                assert max_xy is None
            elif (x is not None) and (y is None):
                assert min_xy == x
                assert max_xy == x
            elif (x is None) and (y is not None):
                assert min_xy == y
                assert max_xy == y
            else:
                assert min_xy == min(x, y)
                assert max_xy == max(x, y)

    @given(engine=sqlite_engines())
    def test_label(self, engine: Engine) -> None:
        table = Table(
            "example",
            MetaData(),
            Column("id", Integer, primary_key=True, autoincrement=True),
            Column("x", Integer),
        )
        ensure_table_created(table, engine)

        sel = select(columnwise_min(table.c.x, table.c.x))
        with engine.begin() as conn:
            _ = conn.execute(sel).all()


class TestCreateEngine:
    @given(temp_path=temp_paths())
    def test_main(self, temp_path: Path) -> None:
        engine = create_engine("sqlite", database=temp_path.name)
        assert isinstance(engine, Engine)


class TestEnsureTableCreated:
    @given(engine=sqlite_engines())
    @mark.parametrize("runs", [param(1), param(2)])
    def test_core(self, engine: Engine, runs: int) -> None:
        table = Table(
            "example", MetaData(), Column("id", Integer, primary_key=True)
        )
        self._run_test(table, engine, runs)

    @given(engine=sqlite_engines())
    @mark.parametrize("runs", [param(1), param(2)])
    def test_orm(self, engine: Engine, runs: int) -> None:
        class Example(cast(Any, declarative_base())):
            __tablename__ = "example"
            id = Column(Integer, primary_key=True)

        self._run_test(Example, engine, runs)

    def _run_test(
        self, table_or_model: Any, engine: Engine, runs: int, /
    ) -> None:
        sel = get_table(table_or_model).select()
        with raises(
            OperationalError, match="no such table"
        ), engine.begin() as conn:
            _ = conn.execute(sel).all()

        for _ in range(runs):
            ensure_table_created(table_or_model, engine)

        with engine.begin() as conn:
            _ = conn.execute(sel).all()


class TestEnsureTableDropped:
    @given(engine=sqlite_engines())
    @mark.parametrize("runs", [param(1), param(2)])
    def test_core(self, engine: Engine, runs: int) -> None:
        table = Table(
            "example", MetaData(), Column("id", Integer, primary_key=True)
        )
        self._run_test(table, engine, runs)

    @given(engine=sqlite_engines())
    @mark.parametrize("runs", [param(1), param(2)])
    def test_orm(self, engine: Engine, runs: int) -> None:
        class Example(cast(Any, declarative_base())):
            __tablename__ = "example"
            id = Column(Integer, primary_key=True)

        self._run_test(Example, engine, runs)

    def _run_test(
        self, table_or_model: Any, engine: Engine, runs: int, /
    ) -> None:
        table = get_table(table_or_model)
        sel = table.select()
        with engine.begin() as conn:
            table.create(conn)
            _ = conn.execute(sel).all()

        for _ in range(runs):
            ensure_table_dropped(table_or_model, engine)

        with raises(
            OperationalError, match="no such table"
        ), engine.begin() as conn:
            _ = conn.execute(sel).all()


class TestGetColumnNames:
    def test_core(self) -> None:
        table = Table(
            "example", MetaData(), Column("id", Integer, primary_key=True)
        )
        self._run_test(table)

    def test_orm(self) -> None:
        class Example(cast(Any, declarative_base())):
            __tablename__ = "example"
            id = Column(Integer, primary_key=True)

        self._run_test(Example)

    def _run_test(self, table_or_model: Any, /) -> None:
        assert get_column_names(table_or_model) == ["id"]


class TestGetColumns:
    def test_core(self) -> None:
        table = Table(
            "example", MetaData(), Column("id", Integer, primary_key=True)
        )
        self._run_test(table)

    def test_orm(self) -> None:
        class Example(cast(Any, declarative_base())):
            __tablename__ = "example"
            id = Column(Integer, primary_key=True)

        self._run_test(Example)

    def _run_test(self, table_or_model: Any, /) -> None:
        columns = get_columns(table_or_model)
        assert isinstance(columns, list)
        assert len(columns) == 1
        assert isinstance(columns[0], Column)


class TestGetTable:
    def test_core(self) -> None:
        table = Table("example", MetaData(), Column("id", primary_key=True))
        assert get_table(table) is table

    def test_orm(self) -> None:
        class Example(cast(Any, declarative_base())):
            __tablename__ = "example"
            id = Column(Integer, primary_key=True)

        table = get_table(Example)
        assert isinstance(table, Table)
        assert table.name == "example"



from pathlib import Path

from pytest import mark
from pytest import param
from pytest import raises

from utilities.atomicwrites import writer


class TestWriter:
    @mark.parametrize(
        "is_binary, contents",
        [
            param(False, "contents", id="text"),
            param(True, b"contents", id="binary"),
        ],
    )
    def test_basic_usage(
        self, tmp_path: Path, is_binary: bool, contents: str | bytes
    ) -> None:
        path = tmp_path.joinpath("file.txt")
        with writer(path) as temp, open(
            temp, mode="wb" if is_binary else "w"
        ) as fh1:
            _ = fh1.write(contents)
        with open(str(path), mode="rb" if is_binary else "r") as fh2:
            assert fh2.read() == contents

    def test_file_exists_error(self, tmp_path: Path) -> None:
        path = tmp_path.joinpath("file.txt")
        with writer(path) as temp1, open(temp1, mode="w") as fh1:
            _ = fh1.write("contents")
        with raises(FileExistsError, match=str(path)), writer(
            path
        ) as temp2, open(temp2, mode="w") as fh2:
            _ = fh2.write("new contents")

    def test_overwrite(self, tmp_path: Path) -> None:
        path = tmp_path.joinpath("file.txt")
        with writer(path) as temp1, open(temp1, mode="w") as fh1:
            _ = fh1.write("contents")
        with writer(path, overwrite=True) as temp2, open(
            temp2, mode="w"
        ) as fh2:
            _ = fh2.write("new contents")
        with open(str(path)) as fh3:
            assert fh3.read() == "new contents"
